import yaml


def state_request_syntax(
    hub, resource_name: str, ref: str, func_name: str, params: dict
):
    example_params = hub.cloudspec.parse.example.extract_params(params)
    request_syntax = {
        f"idem_test_{resource_name}_is_{func_name}": {
            f"{ref}.{func_name}": [
                {key: value} for key, value in example_params.items()
            ]
        }
    }

    return _display(request_syntax)


def extract_params(hub, params: dict):
    result = {}
    for name, param in params.items():
        if param.get("exclude_from_example"):
            continue

        snaked = param.get("snaked")
        param_type = param.get("param_type")
        if param_type == "str":
            param_type = "string"

        if param.get("member", None):
            if param_type == "List[{}]":
                result[snaked] = [
                    hub.cloudspec.parse.example.extract_params(
                        param.get("member", {}).get("params")
                    )
                ]
            elif param_type == "{}":
                result[snaked] = hub.cloudspec.parse.example.extract_params(
                    param.get("member", {}).get("params")
                )
        else:
            if param_type == "list" or param_type == "List[str]":
                result[snaked] = ["value"]
            elif param_type == "Dict[str, str]" or param_type == "Dict[str, {}]":
                result[snaked] = ["key : value"]
            else:
                result[snaked] = param_type

    return result


def _display(data):
    """
    Print the raw data
    Disclaimer: This is taken from https://gitlab.com/vmware/pop/rend/-/blob/master/rend/output/yaml_out.py
    """

    # # yaml safe_dump doesn't know how to represent subclasses of dict.
    # # this registration allows arbitrary dict types to be represented
    # # without conversion to a regular dict.
    def any_dict_representer(dumper, data):
        return dumper.represent_dict(data)

    yaml.add_multi_representer(dict, any_dict_representer, Dumper=yaml.SafeDumper)
    return yaml.safe_dump(data)



"""
Templates for state modules
"""

PRESENT_FUNCTION = """
async def {{function.name}}(
    hub,
    ctx
    {% if function.header_params %}
        {{function.header_params}}
    {% endif -%},
) -> Dict[str, Any]:
    '''{{function.doc|replace("'" * 3, '"' * 3)}}
    Example:
        .. code-block:: sls

          {% if function.example_request_syntax %}
          {{ function.example_request_syntax|indent(12) }}
          {% else %}
              resource_is_{{function.name}}:
                {{ function.ref }}.{{ function.name }}:
                  - {{ function.required_call_params.split(", ")|join("\n                  - ")|replace("=", ": ") }}
          {% endif %}
    '''
"""

ABSENT_FUNCTION = """
async def {{function.name}}(
    hub,
    ctx
    {% if function.header_params %}
        {{function.header_params}}
    {% endif -%},
)  -> Dict[str, Any]:
    '''
    {{function.doc|replace("'" * 3, '"' * 3)}}
    Example:
        .. code-block:: sls

            {% if function.example_request_syntax %}
            {{ function.example_request_syntax|indent(12) }}
            {% else %}
                resource_is_{{function.name}}:
                  {{ function.ref }}.{{ function.name }}:
                    - name: value
                    - resource_id: value
            {% endif %}
    '''
"""

DESCRIBE_FUNCTION = """
async def {{function.name}}(hub, ctx) -> Dict[str, Dict[str, Any]]:
    '''
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    {{function.doc|replace("'" * 3, '"' * 3)}}
    Example:

        .. code-block:: bash

            $ idem describe {{ function.ref }}
    '''
"""



from typing import Dict
from typing import List

from cloudspec import CloudSpecFunction
from cloudspec import CloudSpecParam


def return_type(hub, func_data: CloudSpecFunction):
    ret = ""
    if func_data.return_type:
        ret += "\n\n    Returns:\n"
        ret += f"        {func_data.return_type}"
        ret += "\n"
    return ret


def doc(hub, func_data: CloudSpecFunction):
    if func_data.doc:
        doc = func_data.doc.replace('"""', "'''")
        doc = "\n" + hub.tool.format.wrap.indent(doc, 1) + "\n"
    else:
        doc = ""
    return doc


def parse(hub, functions: Dict[str, CloudSpecFunction], targets: List[str]):
    funcs = {}
    if not functions:
        # It might be an init file
        return funcs
    for function_name in targets:
        function_data = functions.get(function_name)
        # This lets us create partial plugin
        if function_data:
            # These known params should be first in the list of required params according to contract
            idem_known_params = {}
            # Add resource_id parameter
            if function_name in [
                "present",
                "absent",
                "get",
                "create",
                "update",
                "delete",
            ]:
                if (
                    "resource_id"
                    not in hub.cloudspec.parse.param.simple_map(
                        function_data.params
                    ).keys()
                ):
                    idem_known_params["resource_id"] = CloudSpecParam(
                        hub=hub,
                        name="resource_id",
                        doc=f"{function_data.hardcoded.get('resource_name', '').capitalize()} unique ID",
                        required=True
                        if function_name in ["update", "delete", "get"]
                        else False,
                        target="kwargs",
                        target_type="mapping",
                        param_type="str",
                        exclude_from_example=True,
                    )
                else:
                    # if it was already present, we will pop it into known param dict
                    idem_known_params["resource_id"] = function_data.params.pop(
                        "resource_id"
                    )

            # Add name parameter
            if function_name in [
                "present",
                "absent",
                "get",
                "list",
                "create",
                "update",
                "delete",
            ]:
                if (
                    "name"
                    not in hub.cloudspec.parse.param.simple_map(
                        function_data.params
                    ).keys()
                ):
                    idem_known_params["name"] = CloudSpecParam(
                        hub=hub,
                        name="name",
                        doc=f"Idem name of the resource.",
                        required=True
                        if function_name in ["present", "absent"]
                        else False,
                        target="kwargs",
                        target_type="mapping",
                        param_type="str",
                        exclude_from_example=True,
                    )
                else:
                    # if it was already present, we will pop it into known param dict
                    idem_known_params["name"] = function_data.params.pop("name")

            if idem_known_params and function_data.params:
                function_data.params = {**idem_known_params, **function_data.params}

            doc = hub.cloudspec.parse.function.doc(function_data)
            doc += hub.cloudspec.parse.param.sphinx_docs(function_data.params)
            doc += "\n\n    Returns:\n        Dict[str, Any]\n"

            if "absent" == function_name and function_data.params:
                # By idem's contract,
                #   - all absent function parameters except 'name' should not be required
                for param_name, metadata in function_data.params.items():
                    if param_name.lower() != "name":
                        metadata["required"] = False

            param_mapping = hub.cloudspec.parse.param.mappings(function_data.params)

            funcs[function_name] = dict(
                function=dict(
                    name=function_name,
                    hardcoded=function_data.hardcoded,
                    doc=doc,
                    header_params=hub.cloudspec.parse.param.headers(
                        function_data.params
                    ),
                    required_call_params=hub.cloudspec.parse.param.required_call_params(
                        function_data.params
                    ),
                ),
                parameter=dict(
                    mapping=param_mapping,
                ),
            )
    return funcs



from typing import Dict

from cloudspec import CloudSpecParam


def sphinx_docs(hub, parameters: CloudSpecParam) -> str:
    """
    Get the sphinx docs for the parameters
    """
    if not parameters:
        return ""

    ret = "\n    Args:\n"

    # Add required params first
    for name, data in parameters.items():
        if data["required"]:
            ret += add_param_to_sphynx_doc(data)

    # Add not required params after
    for name, data in parameters.items():
        if not data["required"]:
            ret += add_param_to_sphynx_doc(data)

    return ret.rstrip()


def add_param_to_sphynx_doc(param: CloudSpecParam):
    ret = f"        {param.snaked}"
    # Doc strings standard is lower case list[]
    if param.param_type is None:
        param_type = None
    else:
        param_type = param.param_type.replace("list[", "list[")
    if param.member:
        # Complex params are represented as Dict[str, Any] in the sphinx docs
        param_type = param_type.format("dict[str, Any]")

    if param_type:
        ret += f"({param_type}{', Optional' if not param.required else ''})"

    # vRA idem resource provider parsing logic expects parameter doc to be defined on new line
    ret += f":\n            {param.doc}"
    if not ret.endswith("."):
        ret += "."
    if not param.required:
        ret += f" Defaults to {param.default}."

    ret += nested_params_docs(param, "            ")
    ret += "\n\n"
    return ret


def nested_params_docs(param: CloudSpecParam, offset: str):
    """
    Get the sphinx docs for the nested parameters
    """
    ret = ""
    if param.member:
        for nested_param_name, nested_param_data in param.member.params.items():
            # Doc strings standard is lower case list[]
            param_type = nested_param_data.param_type.replace("list[", "list[")
            if nested_param_data.member:
                # This is the next level of nested complex argument.
                param_type = param_type.format("dict[str, Any]")
            # vRA idem resource provider doc parsing logic expects nested properties to be prefixed with *
            ret += f"\n\n{offset}* "

            ret += f"{nested_param_name} ({param_type}{', Optional' if not nested_param_data.required else ''})"
            # vRA idem resource provider doc parsing logic expects parameter doc to be defined on new line
            ret += f":\n{offset}    {nested_param_data.doc}"

            # Recursively add doc for all nested complex arguments
            ret += nested_params_docs(nested_param_data, offset + "    ")
    return ret


def headers(hub, parameters: CloudSpecParam) -> str:
    """
    The arguments that will be put in the function definition
    """
    ret = ""

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_name, param_data in required_params.items():
        ret += add_param_to_header(param_data)

    unrequired_params = {
        name: data for name, data in parameters.items() if not data["required"]
    }

    for param_name, param_data in unrequired_params.items():
        ret += add_param_to_header(param_data)

        # TODO handle this case properly
        if param_data.default == "{}" or param_data.default == "[]":
            ret += f" = None"
        else:
            ret += f" = {param_data.default}"

    return ret


def add_param_to_header(param_data: Dict) -> str:
    ret = f",\n    {param_data.snaked}: "
    if param_data.member:
        # Insert nested complex format for as the argument type.
        # param_type should be "{}|List[{}]|Dict[str, {}]"
        ret += f"{param_data.param_type}".format(
            add_nested_param_to_header(param_data.member, "        ")
        )
    else:
        ret += f"{param_data.param_type}"

    return ret


def add_nested_param_to_header(member: Dict, offset: str) -> str:
    """
    Create a dataclass for a complex nested argument
    """
    ret = f'make_dataclass(\n{offset}"{member.name}",\n{offset}['
    required_params = {
        name: data for name, data in member.get("params").items() if data["required"]
    }

    for param_name, param_data in required_params.items():
        ret += add_nested_param(param_name, param_data, offset)

    unrequired_params = {
        name: data
        for name, data in member.get("params").items()
        if not data["required"]
    }

    for param_name, param_data in unrequired_params.items():
        ret += add_nested_param(param_name, param_data, offset)

    # Remove last ', '
    ret = ret[0:-2]
    ret += f"\n{offset}]\n{offset})"

    return ret


def add_nested_param(param_name: str, param_data, offset: str):
    ret = ""
    ret += f"\n{offset}    "
    # TODO: Nested param name should be snaked
    ret += f'("{param_name}", '
    if param_data.member:
        # Recursively add all nested complex arguments
        ret += param_data.param_type.format(
            add_nested_param_to_header(param_data.member, offset + "    ")
        )
    else:
        ret += f"{param_data.param_type}"

    if param_data["required"]:
        ret += f"), "
    elif param_data.param_type == "int" and param_data.default == 0:
        # Since idem 23.0.3 the default values are populated automatically and value of 0 may fail.
        ret += f", field(default=None)), "
    else:
        ret += f", field(default={param_data.default})), "

    return ret


def required_call_params(hub, parameters: CloudSpecParam) -> str:
    """
    Get a mapping of required function args to the values that will be used in the final caller
    """
    ret = []

    required_params = {
        name: data for name, data in parameters.items() if data["required"]
    }
    for param_data in required_params.values():
        ret.append(f"{param_data.snaked}={param_data.default or 'value'}")

    return ", ".join(ret)


def all_call_params(hub, parameters: CloudSpecParam) -> str:
    """
    Get a mapping of all function args to the values that will be used in the final caller
    """
    ret = []

    all_params = {name: data for name, data in parameters.items()}
    for param_data in all_params.values():
        ret.append(f"{param_data.snaked}={param_data.default or 'value'}")

    return ", ".join(ret)


def mappings(hub, parameters: CloudSpecParam) -> Dict[str, str]:
    ret = {}
    map_params = {
        name: data
        for name, data in parameters.items()
        if data["target_type"] == "mapping"
    }
    for name, data in map_params.items():
        target = data["target"]
        if target not in ret:
            ret[target] = {}
        ret[target][name] = data.snaked

    fmt = lambda item: ", ".join(f'"{k}": {v}' for k, v in item.items())
    return {k: f"{{{fmt(v)}}}" for k, v in ret.items()}


def simple_map(hub, parameters: Dict[str, CloudSpecParam]):
    result = {}
    for k, param_data in parameters.items():
        if k == "name" and param_data.target_type == "arg":
            continue
        result[param_data.snaked] = k
    return result



import pathlib

from cloudspec import CloudSpec


def run(hub, ctx, root_directory: pathlib.Path or str):
    if isinstance(root_directory, str):
        root_directory = pathlib.Path(root_directory)
    cloud_spec = CloudSpec(**ctx.cloud_spec)

    for ref, plugin in cloud_spec.plugins.items():
        for function_name in ["present", "absent"]:
            func_data = plugin.functions.get(function_name)

            if func_data is None:
                continue

            path = root_directory / "example" / function_name
            sls_file = hub.cloudspec.parse.plugin.touch(
                root=path, ref=ref, is_test=False, is_sls=True
            )
            hub.tool.path.touch(sls_file)
            example_request_syntax_output = (
                hub.cloudspec.parse.example.state_request_syntax(
                    resource_name=ref,
                    ref=hub.cloudspec.parse.plugin.mod_ref(ctx, ref, plugin),
                    func_name=function_name,
                    params=func_data.params,
                )
            )

            sls_file.write_text(example_request_syntax_output)



import openapi3


def parse(hub, api: openapi3.OpenAPI):
    schemas = {}

    # Sometimes a schema could just point to different schema object
    # Let's hold them into a separate mapping for adding them into flat schema structure
    transitive_schema_mappings = {}
    # Get a flat structure for schemas

    complex_schema = {}

    # Let's build the reference mapping for complex schema
    for object_name, param in api.components.raw_element.get("schemas").items():
        if "allOf" in param:
            # Get everything as flattened
            complex_schema[object_name] = _resolve_complex_schema(param)
            # Add this in schema mappings so that it can be appended
            transitive_schema_mappings[
                f"#/components/schemas/{object_name}"
            ] = complex_schema[object_name].get("$ref")

    for object_name, param in api.components.raw_element.get("schemas").items():
        parameters = []

        if object_name in complex_schema:
            # Get all the parameters from a resolved 'allOf'
            param = complex_schema[object_name]

        if param.get("$ref", None):
            # It will be resolved later
            transitive_schema_mappings[
                f"#/components/schemas/{object_name}"
            ] = param.get("$ref")

        if param.get("type", "") == "object":
            # "StoreOrder": {
            #     "type": "object",
            #     "properties": {
            #         "name": {
            #             "type": "string",
            #             "description": ""
            #         },
            #     }
            for prop_name, prop_value in param.get("properties", {}).items():
                # Check for nested members first
                # "pet_order": {
                #    "$ref": "#/components/schemas/StoreOrder"
                #  },
                nested_member_ref = prop_value.get("$ref")
                nested_member_ref_type = "nested" if nested_member_ref else None
                if "array" == prop_value.get("type"):
                    # "tags": {
                    #     "type": "array",
                    #     "items": {
                    #         "$ref": "#/definitions/Tag"
                    #     }
                    # },
                    # "photoUrls": {
                    #     "type": "array",
                    #     "items": {
                    #         "type": "string",
                    #     }
                    # },
                    nested_member_ref = prop_value.get("items", {}).get("$ref")
                    # Keeping the default array for now so that it is becomes a list in function params
                    nested_member_ref_type = (
                        "nested_array" if nested_member_ref else "array"
                    )

                param_type = (
                    nested_member_ref_type
                    if nested_member_ref_type
                    else prop_value.get("type")
                )

                parameters.append(
                    dict(
                        idem_name=hub.tool.format.case.snake(prop_name),
                        actual_name=prop_name,
                        required=prop_name in param.get("required", []),
                        type=param_type,
                        member_type=prop_value.get("items", {}).get("type", None)
                        if "array" == param_type
                        else None,
                        nested_member=nested_member_ref,
                        description=prop_value.get("description") or prop_name,
                    )
                )
        else:
            # We may need to sanitize it later
            # "InventoryOrderStatus": {
            #     "type": "string",
            #     "description": "<>",
            # },
            parameters.append(
                dict(
                    idem_name="primitive",
                    type=param.get("type"),
                    description=param.get("description"),
                )
            )

        schemas[f"#/components/schemas/{object_name}"] = parameters

    for key, actual_schema_ref in transitive_schema_mappings.items():
        # Update schemas with actual schema reference
        _, object_name = key.split("/")[-2:]
        if object_name in complex_schema and key in schemas:
            schemas[key].extend(schemas[actual_schema_ref])
        else:
            schemas[key] = schemas[actual_schema_ref]

    return schemas


def _resolve_complex_schema(schema_obj):
    if isinstance(schema_obj, dict):
        if "allOf" in schema_obj:
            resolved_properties = {}
            for schema in schema_obj["allOf"]:
                # resolve allOf for each item in the allOf array
                resolved_properties.update(_resolve_complex_schema(schema))

            # Remove the allOf keyword
            schema_obj.pop("allOf")
            schema_obj.update(resolved_properties)

        # Recursively resolve properties for nested schema objects
        for key, value in schema_obj.items():
            schema_obj[key] = _resolve_complex_schema(value)
    elif isinstance(schema_obj, list):
        schema_obj = [_resolve_complex_schema(item) for item in schema_obj]

    return schema_obj
